// This unit has routines to determine a player's best poker hand

unit PokerHand;

interface

type
  Hand5 = array[1..5] of integer;  // five card poker hand
  Hand7 = array[1..7] of integer;  // two hole cards plus five community cards

// cards are represented by a number from 1 to 52 with 1 being
// the 2 of clubs, and 52 being the ace of spades. Suits within the same
// card value are sorted alphabetically: clubs, diamonds, hearts, spades

const
  CardSuit: array[0..52] of char='xcdhscdhscdhscdhscdhscdhscdhscdhscdhscdhscdhscdhscdhs';
  CardFace: array[0..52] of char='022223333444455556666777788889999TTTTJJJJQQQQKKKKAAAA';
  CardValue: array[0..52] of integer=(0,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,6,6,6,6,
                                      7,7,7,7,8,8,8,8,9,9,9,9,10,10,10,10,11,11,
                                      11,11,12,12,12,12,13,13,13,13,14,14,14,14);

// forward declarations

function StraightFlush(h7: Hand7; var h5: Hand5): boolean;
function Quads(v7: Hand7; var h5: Hand5): boolean;
function FullHouse(v7: Hand7; var h5: Hand5): boolean;
function Flush(h7: Hand7; var h5: Hand5; var s: char): boolean;
function Straight(v7: Hand7; var h5: Hand5): boolean;
function Trips(v7: Hand7; var h5: Hand5): boolean;
function TwoPair(v7: Hand7; var h5: Hand5): boolean;
function Pair(v7: Hand7; var h5: Hand5): boolean;
function HandLong(Rank: integer; Kicker: boolean): string;
function HandRank(h1, h2, b1, b2, b3, b4, b5: integer): integer;

implementation

// Determine if straight flush exists in 7 cards (hole+community)
// If so, return in 5 card hand, passed by reference

function StraightFlush(h7: Hand7; var h5: Hand5): boolean;
var
  s: char;
  i: integer;
  v7: hand7;
begin
  Result:=False;
  if Flush(h7,h5,s)=False then exit;
  for i:=1 to 7 do if CardSuit[h7[i]]=s then v7[i]:=CardValue[h7[i]] else v7[i]:=0;
  Result:=Straight(v7,h5);
end;

// Look for four-of-a-kind

function Quads(v7: Hand7; var h5: Hand5): boolean;
var
  i: integer;
  Count: array[0..14] of integer;
begin
  Result:=False;
  for i:=0 to 14 do Count[i]:=0;
  for i:=1 to 7 do Inc(Count[v7[i]],1);
  for i:=14 downto 2 do if Count[i]=4 then
  begin
    h5[1]:=i;
    h5[2]:=i;
    h5[3]:=i;
    h5[4]:=i;
    if v7[1]<>i then h5[5]:=v7[1] else h5[5]:=v7[5];
    Result:=True;
    Break;
  end;
end;

// Look for full house

function FullHouse(v7: Hand7; var h5: Hand5): boolean;
var
  i,j: integer;
  Count: array[0..14] of integer;
begin
  Result:=False;
  for i:=0 to 14 do Count[i]:=0;
  for i:=1 to 7 do Inc(Count[v7[i]],1);
  j:=0;
  for i:=14 downto 2 do if Count[i]=3 then
  begin
    j:=i;
    Break;
  end;
  if j=0 then Exit;
  for i:=14 downto 2 do if (i<>j) and (Count[i]>=2) then
  begin
    h5[1]:=j;
    h5[2]:=j;
    h5[3]:=j;
    h5[4]:=i;
    h5[5]:=i;
    Result:=True;
    Break;
  end;
end;

// Look for flush

function Flush(h7: Hand7; var h5: Hand5; var s: char): boolean;
var i,j,Clubs,Diamonds,Hearts,Spades: integer;
begin
  Clubs:=0;
  Diamonds:=0;
  Hearts:=0;
  Spades:=0;
  for i:=1 to 7 do
  begin
    s:=CardSuit[h7[i]];
    if s='c' then Inc(Clubs) else
    if s='d' then Inc(Diamonds) else
    if s='h' then Inc(Hearts) else
    if s='s' then Inc(Spades);
  end;
  Result:=False;
  s:='x';
  if Clubs>=5 then s:='c' else
  if Diamonds>=5 then s:='d' else
  if Hearts>=5 then s:='h' else
  if Spades>=5 then s:='s' else exit;
  j:=1;
  for i:=1 to 7 do if CardSuit[h7[i]]=s then
  begin
    h5[j]:=CardValue[h7[i]];
    Inc(j);
    if j>5 then Break;
  end;
  Result:=True;
end;

// Look for a particular card value

function CardInHand(v7: Hand7; v: integer): boolean;
var i: integer;
begin
  Result:=False;
  if v=1 then v:=14;
  for i:=1 to 7 do if v7[i]=v then
  begin
    Result:=True;
    Break;
  end;
end;

// Look for straight

function Straight(v7: Hand7; var h5: Hand5): boolean;
var v: integer;
begin
  Result:=False;
  for v:=14 downto 5 do
  if CardInHand(v7,v) and CardInHand(v7,v-1) and CardInHand(v7,v-2) and
     CardInHand(v7,v-3) and CardInHand(v7,v-4) then
  begin
    h5[1]:=v;
    h5[2]:=v-1;
    h5[3]:=v-2;
    h5[4]:=v-3;
    h5[5]:=v-4;
    Result:=True;
    Break;
  end;
end;

// Look for three-of-a-kind

function Trips(v7: Hand7; var h5: Hand5): boolean;
var
  i,j,k: integer;
  Count: array[0..14] of integer;
begin
  Result:=False;
  for i:=0 to 14 do Count[i]:=0;
  for i:=1 to 7 do Inc(Count[v7[i]],1);
  for i:=14 downto 2 do if Count[i]=3 then
  begin
    h5[1]:=i;
    h5[2]:=i;
    h5[3]:=i;
    j:=4;
    for k:=1 to 7 do if v7[k]<>i then
    begin
      h5[j]:=v7[k];
      inc(j);
      if j>5 then Break;
    end;
    Result:=True;
    Break;
  end;
end;

// Look for two pair

function TwoPair(v7: Hand7; var h5: Hand5): boolean;
var
  i,j,k: integer;
  Count: array[0..14] of integer;
begin
  Result:=False;
  for i:=0 to 14 do Count[i]:=0;
  for i:=1 to 7 do Inc(Count[v7[i]],1);
  j:=1;
  for i:=14 downto 2 do if Count[i]=2 then
  begin
    h5[j]:=i;
    h5[j+1]:=i;
    Inc(j,2);
    if j=5 then
    begin
      for k:=1 to 7 do if (v7[k]<>h5[1]) and (v7[k]<>h5[3]) then
      begin
        h5[5]:=v7[k];
        Break;
      end;
      Result:=True;
      Break;
    end;
  end;
end;

// Look for a pair

function Pair(v7: Hand7; var h5: Hand5): boolean;
var
  i,j,k: integer;
  Count: array[0..14] of integer;
begin
  Result:=False;
  for i:=0 to 14 do Count[i]:=0;
  for i:=1 to 7 do Inc(Count[v7[i]],1);
  for i:=14 downto 2 do if Count[i]=2 then
  begin
    h5[1]:=i;
    h5[2]:=i;
    j:=3;
    for k:=1 to 7 do if v7[k]<>i then
    begin
      h5[j]:=v7[k];
      Inc(j);
      if j>5 then Break;
    end;
    Result:=True;
    Break;
  end;
end;

function HandLong(Rank: integer; Kicker: boolean): string;
const
  sname: array[0..14] of string=('','Ace','Deuce','Three','Four','Five','Six',
    'Seven','Eight','Nine','Ten','Jack','Queen','King','Ace');
  pname: array[0..14] of string=('','Aces','Deuces','Threes','Fours','Fives','Sixes',
    'Sevens','Eights','Nines','Tens','Jacks','Queens','Kings','Aces');
  cname: array[0..14] of string=('','A','2','3','4','5','6','7','8','9','T','J',
    'Q','K','A');
var
  s: string;
  k1,k2,k3,k4,k5: integer;
begin
  k1:=(Rank and $0f0000) shr 16;
  k2:=(Rank and $00f000) shr 12;
  k3:=(Rank and $000f00) shr 8;
  k4:=(Rank and $0000f0) shr 4;
  k5:=Rank and $00000f;
  if Rank= $9EDCBA then s:='Royal Flush' else
  if Rank>=$900000 then s:='Straight Flush, '+sname[k5]+' to '+sname[k1] else
  if Rank>=$800000 then
  begin
    s:='Four of a Kind, '+pname[k1];
    if Kicker then s:=s+' +'+cname[k5];
  end else
  if Rank>=$700000 then s:='Full House, '+pname[k1]+' full of '+pname[k4] else
  if Rank>=$600000 then
  begin
    s:='Flush, '+sname[k1]+' high';
    if Kicker then s:=s+' +'+cname[k2]+cname[k3]+cname[k4]+cname[k5];
  end else
  if Rank>=$500000 then s:='Straight, '+sname[k5]+' to '+sname[k1] else
  if Rank>=$400000 then
  begin
    s:='Three of a Kind, '+pname[k1];
    if Kicker then s:=s+' +'+cname[k4]+cname[k5];
  end else
  if Rank>=$300000 then
  begin
    s:='Two Pair, '+pname[k1]+' and '+pname[k3];
    if Kicker then s:=s+' +'+cname[k5];
  end else
  if Rank>=$200000 then
  begin
    s:='Pair of '+pname[k1];
    if Kicker then s:=s+' +'+cname[k3]+cname[k4]+cname[k5];
  end else
  if Rank>=$100000 then
  begin
    s:='High Card '+sname[k1];
    if Kicker then s:=s+' +'+cname[k2]+cname[k3]+cname[k4]+cname[k5];
  end
  else s:='Error';
  Result:=s;
end;

function HandRank(h1, h2, b1, b2, b3, b4, b5: integer): integer;
var
  i, j, k, t: integer;
  s: char;
  h5: hand5;
  h7, v7: hand7;
begin
  Result := 0;
  if (h1 = 0) or (h2 = 0) then exit;
  // combine player and board cards
  h7[1] := b1;
  h7[2] := b2;
  h7[3] := b3;
  h7[4] := b4;
  h7[5] := b5;
  h7[6] := h1;
  h7[7] := h2;
  // sort hand
  for j := 1 to 6 do
    for k := j + 1 to 7 do
    if h7[k] > h7[j] then
    begin
      t := h7[k];
      h7[k] := h7[j];
      h7[j] := t;
    end;
  for i := 1 to 7 do v7[i] := CardValue[h7[i]];
  if StraightFlush(h7, h5) then Result := $900000 else
  if Quads(v7, h5) then Result := $800000 else
  if FullHouse(v7, h5) then Result := $700000 else
  if Flush(h7, h5, s) then Result := $600000 else
  if Straight(v7, h5) then Result := $500000 else
  if Trips(v7, h5) then Result := $400000 else
  if TwoPair(v7, h5) then Result := $300000 else
  if Pair(v7, h5) then Result := $200000 else
  begin  // high card
    Result := $100000;
    for i := 1 to 5 do h5[i] := CardValue[h7[i]];
  end;
  Result := Result + (h5[1] shl 16) + (h5[2] shl 12) + (h5[3] shl 8) + (h5[4] shl 4) + h5[5];
end;

end.
