// This unit calculates a secure hash using SHA-256.
// Hash is created as an object so multiple hashes can
// be calculated from multiple threads at one time.

unit SHA256;

interface
type

  TSHAObject = class(TObject)
  private
    H0,H1,H2,H3,H4,H5,H6,H7: longword;
    P: integer;
    Total: int64;
    IBuf: array[0..63] of byte;
    function BigEndian(l: longword): longword;
    function Ch(x,y,z: longword): longword;
    function Maj(x,y,z: longword): longword;
    function Sum0(x: longword): longword;
    function Sum1(x: longword): longword;
    function Sigma0(x: longword): longword;
    function Sigma1(x: longword): longword;
    procedure BSwap(i,j: integer);
    procedure Main;
  public
    Hash: array[0..31] of byte;
    constructor Create;
    procedure HashBuffer(Buffer: pointer; Size: integer);
    procedure FinalizeHash;
  end;

implementation

constructor TSHAObject.Create;
begin
  inherited Create;
  H0:=$6a09e667;
  H1:=$bb67ae85;
  H2:=$3c6ef372;
  H3:=$a54ff53a;
  H4:=$510e527f;
  H5:=$9b05688c;
  H6:=$1f83d9ab;
  H7:=$5be0cd19;
  FillChar(Hash,32,0);
  P:=0;
  Total:=0;
end;

procedure TSHAObject.HashBuffer(Buffer: pointer; Size: integer);
type
  TBuf = array[0..MaxInt-1] of byte;
var
  x,z: integer;
  Buf: ^TBuf;
begin
  Buf:=Buffer;
  Inc(Total,Size);
  z:=0;
  while Size>0 do
  begin
    x:=64-P;
    if x>Size then x:=Size;
    move(Buf^[z],IBuf[P],x);
    z:=z+x;
    P:=P+x;
    if P=64 then
    begin
      Main;
      P:=0;
    end;
    Dec(Size,x);
  end;
end;

procedure TSHAObject.FinalizeHash;
begin
  IBuf[P]:=$80;
  Inc(P);
  while P<56 do
  begin
    IBuf[P]:=0;
    Inc(P);
  end;
  if P>56 then
  begin
    while P<64 do
    begin
      IBuf[P]:=0;
      Inc(P);
    end;
    Main;
    FillChar(IBuf,64,0);
  end;
  Total:=Total*8;
  Move(Total,IBuf[56],8);
  BSwap(56,63);
  BSwap(57,62);
  BSwap(58,61);
  BSwap(59,60);
  Main;
  H0:=BigEndian(H0);
  H1:=BigEndian(H1);
  H2:=BigEndian(H2);
  H3:=BigEndian(H3);
  H4:=BigEndian(H4);
  H5:=BigEndian(H5);
  H6:=BigEndian(H6);
  H7:=BigEndian(H7);
  Move(H0,Hash[0],4);
  Move(H1,Hash[4],4);
  Move(H2,Hash[8],4);
  Move(H3,Hash[12],4);
  Move(H4,Hash[16],4);
  Move(H5,Hash[20],4);
  Move(H6,Hash[24],4);
  Move(H7,Hash[28],4);
end;

function TSHAObject.BigEndian(l: longword): longword;
var a,b,c,d: byte;
begin
  a:=l;
  b:=l shr 8;
  c:=l shr 16;
  d:=l shr 24;
  Result:=d or (c shl 8) or (b shl 16) or (a shl 24);
end;

function TSHAObject.Ch(x,y,z: longword): longword;
begin
  result:=(x and y) xor (not(x) and z);
end;

function TSHAObject.Maj(x,y,z: longword): longword;
begin
  result:=(x and y) xor (x and z) xor (y and z);
end;

function TSHAObject.Sum0(x: longword): longword;
begin
  result:=((x shr 2) or (x shl 30)) xor ((x shr 13) or (x shl 19)) xor
          ((x shr 22) or (x shl 10));
end;

function TSHAObject.Sum1(x: longword): longword;
begin
  result:=((x shr 6) or (x shl 26)) xor ((x shr 11) or (x shl 21)) xor
          ((x shr 25) or (x shl 7));
end;

function TSHAObject.Sigma0(x: longword): longword;
begin
  result:=((x shr 7) or (x shl 25)) xor ((x shr 18) or (x shl 14)) xor (x shr 3);
end;

function TSHAObject.Sigma1(x: longword): longword;
begin
  result:=((x shr 17) or (x shl 15)) xor ((x shr 19) or (x shl 13)) xor (x shr 10);
end;

procedure TSHAObject.BSwap(i,j: integer);
var b: byte;
begin
  b:=IBuf[i];
  IBuf[i]:=IBuf[j];
  IBuf[j]:=b;
end;

procedure TSHAObject.Main;
const
  k: array[0..63] of longword=(
  $428a2f98,$71374491,$b5c0fbcf,$e9b5dba5,$3956c25b,$59f111f1,$923f82a4,$ab1c5ed5,
  $d807aa98,$12835b01,$243185be,$550c7dc3,$72be5d74,$80deb1fe,$9bdc06a7,$c19bf174,
  $e49b69c1,$efbe4786,$0fc19dc6,$240ca1cc,$2de92c6f,$4a7484aa,$5cb0a9dc,$76f988da,
  $983e5152,$a831c66d,$b00327c8,$bf597fc7,$c6e00bf3,$d5a79147,$06ca6351,$14292967,
  $27b70a85,$2e1b2138,$4d2c6dfc,$53380d13,$650a7354,$766a0abb,$81c2c92e,$92722c85,
  $a2bfe8a1,$a81a664b,$c24b8b70,$c76c51a3,$d192e819,$d6990624,$f40e3585,$106aa070,
  $19a4c116,$1e376c08,$2748774c,$34b0bcb5,$391c0cb3,$4ed8aa4a,$5b9cca4f,$682e6ff3,
  $748f82ee,$78a5636f,$84c87814,$8cc70208,$90befffa,$a4506ceb,$bef9a3f7,$c67178f2);
var
  t: integer;
  a,b,c,d,e,f,g,h,t1,t2: longword;
  w: array[0..63] of longword;
begin
  Move(IBuf,w,64);
  for t:=0 to 15 do w[t]:=BigEndian(w[t]);
  for t:=16 to 63 do w[t]:=Sigma1(w[t-2])+w[t-7]+Sigma0(w[t-15])+w[t-16];
  a:=H0;
  b:=H1;
  c:=H2;
  d:=H3;
  e:=H4;
  f:=H5;
  g:=H6;
  h:=H7;
  for t:=0 to 63 do
  begin
    t1:=h+Sum1(e)+Ch(e,f,g)+k[t]+w[t];
    t2:=Sum0(a)+Maj(a,b,c);
    h:=g;
    g:=f;
    f:=e;
    e:=d+t1;
    d:=c;
    c:=b;
    b:=a;
    a:=t1+t2;
  end;
  H0:=H0+a;
  H1:=H1+b;
  H2:=H2+c;
  H3:=H3+d;
  H4:=H4+e;
  H5:=H5+f;
  H6:=H6+g;
  H7:=H7+h;
end;

end.
